# This file is part of Tulip (http://tulip.labri.fr)
#
# Authors: David Auber and the Tulip development Team
# from LaBRI, University of Bordeaux
#
# Tulip is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Tulip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.

"""
This module provides utility functions to register Tulip plugins written
in Python in the plugins database.
"""

import sys
import traceback
if sys.version_info[0] == 3:
    from imp import reload # noqa

from tulip import tlp # noqa

pluginFactory = {}
pluginModules = {}
testMode = False


def setTestMode(mode):
    global testMode
    testMode = mode


def getCallingModuleName():
    import sys
    if sys.version_info[0] == 3:
        frames = list(sys._current_frames().values())
    else:
        frames = sys._current_frames().values()
    for i in range(len(frames)):
        f = frames[i]
        if f.f_globals['__name__'] == 'tulipplugins':
            f = f.f_back
            break
    while f.f_globals['__name__'] == 'tulipplugins':
        f = f.f_back
    return f.f_globals['__name__']


def reloadTulipPythonPlugin(pluginName):
    if pluginName in pluginModules:
        module = pluginModules[pluginName]
        code = 'import %s\n' % module
        code += 'reload(%s)\n' % module
        exec(code)


def reloadTulipPythonPlugins():
    for plugin in pluginModules.keys():
        reloadTulipPythonPlugin(plugin)


def removePlugin(pluginName):
    tulipUtilsOk = True
    try:
        import tuliputils
    except ImportError:
        tulipUtilsOk = False
    if tulipUtilsOk:
        tuliputils.removePlugin(pluginName)


def initFactory(self):
    tlp.PluginFactory.__init__(self)
    self.registerFactory()


def runPlugin(plugin):
    ret = False
    try:
        ret = plugin.real_run()
    except Exception:
        if plugin.pluginProgress:
            plugin.pluginProgress.setError(traceback.format_exc())
        # Case where the plugin execution has not been launched through the
        # Tulip GUI, so print the stack trace to stderr
        if type(plugin.pluginProgress) == tlp.SimplePluginProgress:
            sys.stdout.write(('There was an error when running Python plugin '
                              'named "%s". See stack trace below.\n') %
                             plugin.name())
            traceback.print_exc()
    return ret


def importGraph(plugin):
    ret = False
    try:
        ret = plugin.real_importGraph()
    except Exception:
        if plugin.pluginProgress:
            plugin.pluginProgress.setError(traceback.format_exc())
        # Case where the plugin execution has not been launched through
        # the Tulip GUI, so print the stack trace to stderr
        if type(plugin.pluginProgress) == tlp.SimplePluginProgress:
            sys.stdout.write(('There was an error when running Python plugin '
                              'named "%s". See stack trace below.\n') %
                             plugin.name())
            traceback.print_exc()
    return ret


def exportGraph(plugin, os):
    ret = False
    try:
        ret = plugin.real_exportGraph(os)
    except Exception:
        if plugin.pluginProgress:
            plugin.pluginProgress.setError(traceback.format_exc())
        # Case where the plugin execution has not been launched through
        # the Tulip GUI, so print the stack trace to stderr
        if type(plugin.pluginProgress) == tlp.SimplePluginProgress:
            sys.stdout.write(('There was an error when running Python plugin '
                              'named "%s". See stack trace below.\n') %
                             plugin.name())
            traceback.print_exc()
    return ret


def createPlugin(context, pluginModule, pluginClassName, pluginName,
                 author, date, info, release, group):
    plugin = eval('%s.%s(context)' % (pluginModule, pluginClassName),
                  globals(), locals())
    if hasattr(plugin, 'run'):
        plugin.real_run = plugin.run
        plugin.run = lambda: runPlugin(plugin)
    elif hasattr(plugin, 'importGraph'):
        plugin.real_importGraph = plugin.importGraph
        plugin.importGraph = lambda: importGraph(plugin)
    elif hasattr(plugin, 'exportGraph'):
        plugin.real_exportGraph = plugin.exportGraph
        plugin.exportGraph = lambda os: exportGraph(plugin, os)
    plugin.name = lambda: pluginName
    plugin.group = lambda: group
    plugin.author = lambda: author
    plugin.date = lambda: date
    plugin.info = lambda: info
    plugin.release = lambda: release
    plugin.tulipRelease = lambda: tlp.getTulipRelease()
    plugin.programmingLanguage = lambda: 'Python'
    return plugin


def registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group):
    """
    tulipplugins.registerPluginOfGroup(pluginClassName, pluginName, author, date, info, release, group)

    Registers a Tulip plugin written in Python in the plugins database and
    inserts it in a specific group.

    :param pluginClassName: the name of the Python class implementing the plugin
    :type pluginClassName: string
    :param pluginName: the name of the plugin as it will appear in the interface
        and that will be used to call it
    :type pluginName: string
    :param author: the name of the plugin's author
    :type author: string
    :param date: the date of creation of the plugin
    :type date: string
    :param info: some information relative to the plugin
    :type info: string
    :param release: the version number of the plugin in the form X.Y
    :type release: string
    :param group: the name of the group in which the plugin will be inserted
    :type group: string
    """ # noqa

    global pluginFactory
    removePlugin(pluginName)
    pluginModule = getCallingModuleName()
    try:
        globals()[pluginModule] = __import__(pluginModule, globals(), locals(),
                                             [], 0)
        eval('%s.%s' % (pluginModule, pluginClassName), globals(), locals())
        if testMode:
            return
        pluginModules[pluginName] = pluginModule
        pluginFactory[pluginName] = type(
            '%sFactory' % pluginClassName, (tlp.PluginFactory,),
            {
                '__init__': (lambda self: initFactory(self)),
                'createPluginObject': (lambda self, context: createPlugin(
                    context, pluginModule, pluginClassName, pluginName, author,
                    date, info, release, group))
            }
        )()

    except Exception:
        sys.stdout.write(('There was an error when registering Python plugin '
                          'named "%s". See stack trace below.\n') %
                         pluginName)
        raise


def registerPlugin(pluginClassName, pluginName, author, date, info, release):
    """
    tulipplugins.registerPlugin(pluginClassName, pluginName, author, date, info, release)

    Registers a Tulip plugin written in Python in the plugins database.

    :param pluginClassName: the name of the Python class implementing
        the plugin
    :type pluginClassName: string
    :param pluginName: the name of the plugin as it will appear in the
        interface and that will be used to call it
    :type pluginName: string
    :param author: the name of the plugin's author
    :type author: string
    :param date: the date of creation of the plugin
    :type date: string
    :param info: some information relative to the plugin
    :type info: string
    :param release: the version number of the plugin in the form X.Y
    :type release: string
    """ # noqa

    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


# For backward compatibility with Tulip 3.x
def registerAlgorithmPlugin(pluginClassName, pluginName, author, date, info,
                            release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerAlgorithmPluginOfGroup(pluginClassName, pluginName, author, date,
                                   info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerLayoutPlugin(pluginClassName, pluginName, author, date, info,
                         release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerLayoutPluginOfGroup(pluginClassName, pluginName, author, date,
                                info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerDoublePlugin(pluginClassName, pluginName, author, date, info,
                         release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerDoublePluginOfGroup(pluginClassName, pluginName, author, date,
                                info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerIntegerPlugin(pluginClassName, pluginName, author, date, info,
                          release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerIntegerPluginOfGroup(pluginClassName, pluginName, author, date,
                                 info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerBooleanPlugin(pluginClassName, pluginName, author, date, info,
                          release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerBooleanPluginOfGroup(pluginClassName, pluginName, author, date,
                                 info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerSizePlugin(pluginClassName, pluginName, author, date, info,
                       release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerSizePluginOfGroup(pluginClassName, pluginName, author, date, info,
                              release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerColorPlugin(pluginClassName, pluginName, author, date, info,
                        release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerColorPluginOfGroup(pluginClassName, pluginName, author, date, info,
                               release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerImportPlugin(pluginClassName, pluginName, author, date, info,
                         release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerImportPluginOfGroup(pluginClassName, pluginName, author, date,
                                info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)


def registerExportPlugin(pluginClassName, pluginName, author, date, info,
                         release):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, '')


def registerExportPluginOfGroup(pluginClassName, pluginName, author, date,
                                info, release, group):
    registerPluginOfGroup(pluginClassName, pluginName, author, date, info,
                          release, group)



# This file is part of Tulip (http://tulip.labri.fr)
#
# Authors: David Auber and the Tulip development Team
# from LaBRI, University of Bordeaux
#
# Tulip is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Tulip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.

# -*- coding: utf-8 -*-

import os
import os.path
import sys
import traceback
import platform

_tulipNativeLibsPath = os.path.join(os.path.dirname(__file__), 'native')
sys.path.append(_tulipNativeLibsPath)

if platform.system() == 'Windows':
    os.environ['PATH'] = '%s;%s;%s' % (
        _tulipNativeLibsPath,
        os.path.join(_tulipNativeLibsPath, '../../..'),
        os.environ['PATH'])

import _tulip # noqa

sys.path.pop()


class tlpType(_tulip.tlp.__class__):

    def __getattr__(cls, name):
        if hasattr(_tulip.tlp, name):
            return _tulip.tlp.getTulipGlobalVar(name)
        else:
            raise AttributeError(name)

    def __setattr__(cls, name, value):
        if hasattr(_tulip.tlp, name):
            _tulip.tlp.setTulipGlobalVar(name, value)
        else:
            super(tlpType, cls).__setattr__(name, value)


# utility function from the 'six' module
def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

    return type.__new__(metaclass, 'temporary_class', (), {})


class tlp(with_metaclass(tlpType, _tulip.tlp)):

    @staticmethod
    def loadTulipPythonPlugin(pluginFilePath):
        if not os.path.isfile(pluginFilePath):
            print('[tulip] Error: Path %s is not a valid file' %
                  pluginFilePath)
            return False

        try:
            pluginFile = open(pluginFilePath)
            pluginFileContent = pluginFile.read()
        except Exception:
            return False

        if 'tulipplugins.register' not in pluginFileContent:
            return False

        modulePath = os.path.dirname(pluginFilePath)
        moduleName = os.path.basename(pluginFilePath)[:-3]

        if modulePath not in sys.path:
            sys.path.append(modulePath)

        try:
            __import__(moduleName)
        except ImportError:
            sys.stdout.write(('There was an error when trying to load the'
                              ' Tulip Python plugin from the file %s. See '
                              'stack trace below.\\n') % pluginFilePath)
            traceback.print_exc()
            return False

        return True

    @staticmethod
    def loadTulipPluginsFromDir(pluginsDirPath, loadCppPlugin=True,
                                pluginLoader=None):
        if not os.path.exists(pluginsDirPath):
            return False

        if loadCppPlugin:
            tlp.loadPluginsFromDir(pluginsDirPath, pluginLoader, False)

        files = os.listdir(pluginsDirPath)

        for file in files:
            filePath = os.path.join(pluginsDirPath, file)
            if not os.path.isdir(filePath) and filePath.endswith('.py'):
                tlp.loadTulipPythonPlugin(filePath)

        for file in files:
            filePath = os.path.join(pluginsDirPath, file)
            if os.path.isdir(filePath):
                tlp.loadTulipPluginsFromDir(filePath, loadCppPlugin,
                                            pluginLoader)

        return True


tulipVersion = tlp.getTulipRelease()
tulipVersion = tulipVersion[:tulipVersion.rfind('.')]

startupScriptsPath = os.path.join(
    tlp.TulipLibDir, 'tulip/python/startup')
startupScriptsHomePath = os.path.join(
    os.path.expanduser('~'), '.Tulip-%s/python/startup' % tulipVersion)


def runStartupScripts(scriptsPath):
    if not os.path.exists(scriptsPath):
        return

    files = os.listdir(scriptsPath)

    for file in files:
        filePath = os.path.join(scriptsPath, file)
        if os.path.isfile(filePath) and filePath.endswith('.py'):
            exec(compile(open(filePath).read(), filePath, 'exec'),
                 globals(), locals())


runStartupScripts(startupScriptsPath)
runStartupScripts(startupScriptsHomePath)

tlpPythonPluginsPath = os.path.join(
    tlp.TulipLibDir, 'tulip/python/tulip/plugins')
tlpPythonPluginsHomePath = os.path.join(
    os.path.expanduser('~'), '.Tulip-%s/plugins/python' % tulipVersion)

tlp.loadTulipPluginsFromDir(tlpPythonPluginsPath, False)
tlp.loadTulipPluginsFromDir(tlpPythonPluginsHomePath, False)

_tulipNativePluginsPath = os.path.join(_tulipNativeLibsPath, 'plugins')

# fix loading of Tulip plugins when the tulip module has been
# installed with the pip tool
if platform.system() == 'Linux' and os.path.exists(_tulipNativePluginsPath):
    dlOpenFlagsBackup = sys.getdlopenflags()
    if sys.version_info < (3, 6):
        import DLFCN
        dlOpenFlags = DLFCN.RTLD_NOW | DLFCN.RTLD_GLOBAL
    else:
        dlOpenFlags = os.RTLD_NOW | os.RTLD_GLOBAL
    sys.setdlopenflags(dlOpenFlags)

tlp.loadTulipPluginsFromDir(_tulipNativePluginsPath)

# load bundled Tulip Python plugins when the tulip module has been
# installed with the pip tool
if not sys.argv[0] == 'tulip':
    tlp.loadTulipPluginsFromDir(
        os.path.join(os.path.dirname(__file__), 'plugins'))

if platform.system() == 'Linux' and os.path.exists(_tulipNativePluginsPath):
    sys.setdlopenflags(dlOpenFlagsBackup)

__all__ = ['tlp']
__author__ = 'David Auber and the Tulip development Team'
__license__ = 'LGPLv3'
__version__ = tlp.getTulipRelease()
__email__ = 'tulipdev@labri.fr'
__status__ = 'Production'



# This file is part of Tulip (http://tulip.labri.fr)
#
# Authors: David Auber and the Tulip development Team
# from LaBRI, University of Bordeaux
#
# Tulip is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Tulip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.

import unittest

from tulip import tlp

NB_NODES = 5
NB_EDGES = 2 * (NB_NODES - 1)


class TestGraphStructure(unittest.TestCase):

    # generate a simple graph with a cycle going from node 0
    # to node (NB_NODES-2) and node (NB_NODES-1) connected to all the others
    def setUp(self):
        self.graph = tlp.newGraph()
        self.nodes = []
        self.edges = []
        for i in range(NB_NODES):
            self.nodes.append(self.graph.addNode())

        for i in range(NB_NODES - 1):
            if i != NB_NODES - 2:
                self.edges.append(
                    self.graph.addEdge(self.nodes[i], self.nodes[i+1]))
            else:
                self.edges.append(
                    self.graph.addEdge(self.nodes[i], self.nodes[0]))
            self.edges.append(
                self.graph.addEdge(self.nodes[i], self.nodes[-1]))

    def tearDown(self):
        self.graph = None
        self.nodes = None
        self.edges = None

    def test_nb_elements(self):
        self.assertEqual(self.graph.numberOfNodes(), NB_NODES)
        self.assertEqual(self.graph.numberOfEdges(), NB_EDGES)
        self.assertEqual(len(list(self.graph.getNodes())), NB_NODES)
        self.assertEqual(len(list(self.graph.getEdges())), NB_EDGES)
        self.assertEqual(len(self.graph.nodes()), NB_NODES)
        self.assertEqual(len(self.graph.edges()), NB_EDGES)

    def test_is_element(self):
        for n in self.nodes:
            self.assertTrue(self.graph.isElement(n))
            self.assertTrue(n in self.graph)

        for e in self.edges:
            self.assertTrue(self.graph.isElement(e))
            self.assertTrue(e in self.graph)

        for n in self.graph.getNodes():
            self.assertIn(n, self.nodes)

        for n in self.graph.nodes():
            self.assertIn(n, self.nodes)

        for e in self.graph.getEdges():
            self.assertIn(e, self.edges)
            self.assertIn(self.graph.source(e), self.nodes)
            self.assertIn(self.graph.target(e), self.nodes)
            ends = self.graph.ends(e)
            self.assertIn(ends[0], self.nodes)
            self.assertIn(ends[1], self.nodes)

        for e in self.graph.edges():
            self.assertIn(e, self.edges)
            self.assertIn(self.graph.source(e), self.nodes)
            self.assertIn(self.graph.target(e), self.nodes)

        for i in range(NB_NODES - 1):
            self.assertTrue(self.graph.hasEdge(self.nodes[i], self.nodes[i+1]))
            self.assertIn(self.graph.existEdge(self.nodes[i], self.nodes[i+1]),
                          self.edges)
            self.assertTrue((self.nodes[i], self.nodes[i+1]) in self.graph)
            self.assertTrue(self.graph.hasEdge(self.nodes[i], self.nodes[-1]))
            self.assertIn(self.graph.existEdge(self.nodes[i], self.nodes[-1]),
                          self.edges)
            self.assertTrue((self.nodes[i], self.nodes[-1]) in self.graph)

        self.assertFalse(self.graph.isElement(tlp.node(NB_NODES)))
        self.assertTrue(tlp.node(NB_NODES) not in self.graph)
        self.assertFalse(self.graph.isElement(tlp.edge(NB_EDGES)))
        self.assertTrue(tlp.edge(NB_EDGES) not in self.graph)

    def test_node_degrees(self):
        for i in range(NB_NODES-1):
            self.assertEqual(self.graph.deg(self.nodes[i]), 3)
            self.assertEqual(self.graph.indeg(self.nodes[i]), 1)
            self.assertEqual(self.graph.outdeg(self.nodes[i]), 2)

        self.assertEqual(self.graph.deg(self.nodes[NB_NODES-1]), NB_NODES-1)
        self.assertEqual(self.graph.indeg(self.nodes[NB_NODES-1]), NB_NODES-1)
        self.assertEqual(self.graph.outdeg(self.nodes[NB_NODES-1]), 0)

    def test_add_invalid_edge(self):
        with self.assertRaises(Exception) as cm:
            self.graph.addEdge(tlp.node(NB_NODES), tlp.node(NB_NODES+1))
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

    def test_del_elements(self):
        with self.assertRaises(Exception) as cm:
            self.graph.delNode(tlp.node(NB_NODES))
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.delEdge(tlp.edge(NB_EDGES))
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        for i, e in enumerate(self.edges):
            self.graph.delEdge(e)
            self.assertEqual(self.graph.numberOfEdges(), NB_EDGES - (i+1))
            self.assertFalse(self.graph.isElement(e))

        for i, n in enumerate(self.nodes):
            self.graph.delNode(n)
            self.assertEqual(self.graph.numberOfNodes(), NB_NODES - (i+1))
            self.assertFalse(self.graph.isElement(n))

    def test_del_nodes_from_list(self):
        invalid_nodes = [tlp.node(NB_NODES), tlp.node(NB_NODES+1)]
        with self.assertRaises(Exception) as cm:
            self.graph.delNodes(invalid_nodes)
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        self.graph.delNodes(self.nodes)
        self.assertEqual(self.graph.numberOfNodes(), 0)
        self.assertEqual(self.graph.numberOfEdges(), 0)

    def test_del_nodes_from_iterator(self):
        self.graph.delNodes(self.graph.getNodes())
        self.assertEqual(self.graph.numberOfNodes(), 0)
        self.assertEqual(self.graph.numberOfEdges(), 0)

    def test_del_edges_from_list(self):
        invalid_edges = [tlp.edge(NB_EDGES), tlp.edge(NB_EDGES+1)]
        with self.assertRaises(Exception) as cm:
            self.graph.delEdges(invalid_edges)
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        self.graph.delEdges(self.edges)
        self.assertEqual(self.graph.numberOfEdges(), 0)

    def test_del_edges_from_iterator(self):
        self.graph.delEdges(self.graph.getEdges())
        self.assertEqual(self.graph.numberOfEdges(), 0)

    def test_add_then_del_nodes(self):
        new_nodes = self.graph.addNodes(NB_NODES)
        self.assertEqual(self.graph.numberOfNodes(), 2*NB_NODES)

        for n in new_nodes:
            self.assertTrue(self.graph.isElement(n))

        self.graph.delNodes(new_nodes)
        self.assertEqual(self.graph.numberOfNodes(), NB_NODES)

        for n in new_nodes:
            self.assertFalse(self.graph.isElement(n))

    def test_add_then_del_edges(self):
        invalid_edges = [(tlp.node(NB_NODES), tlp.node(NB_NODES+1)),
                         (tlp.node(NB_NODES+1), tlp.node(NB_NODES+2))]

        with self.assertRaises(Exception) as cm:
            self.graph.addEdges(invalid_edges)
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        edges_to_add = []
        for i in range(NB_EDGES):
            edge = (self.graph.getRandomNode(), self.graph.getRandomNode())
            edges_to_add.append(edge)

        new_edges = self.graph.addEdges(edges_to_add)
        self.assertEqual(self.graph.numberOfEdges(), 2*NB_EDGES)

        for i, e in enumerate(new_edges):
            self.assertTrue(self.graph.isElement(e))
            self.assertEqual(self.graph.source(e), edges_to_add[i][0])
            self.assertEqual(self.graph.target(e), edges_to_add[i][1])

        self.graph.delEdges(new_edges)
        self.assertEqual(self.graph.numberOfEdges(), NB_EDGES)

        for n in new_edges:
            self.assertFalse(self.graph.isElement(e))

    def test_add_invalid_node_or_edge_to_root_graph(self):
        with self.assertRaises(Exception) as cm:
            self.graph.addNode(tlp.node(NB_NODES))
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.addEdge(tlp.edge(NB_EDGES))
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

    def test_edge_orders(self):
        with self.assertRaises(Exception) as cm:
            self.graph.setEdgeOrder(tlp.node(NB_NODES), [])
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.swapEdgeOrder(tlp.node(NB_NODES),
                                     tlp.edge(), tlp.edge())
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.swapEdgeOrder(self.graph.getRandomNode(),
                                     tlp.edge(NB_EDGES), tlp.edge(NB_EDGES+1))
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        for n in self.graph.getNodes():
            edges = list(self.graph.getInOutEdges(n))
            edges = list(reversed(edges))
            self.graph.setEdgeOrder(n, edges)
            self.assertEqual(list(self.graph.getInOutEdges(n)), edges)
            self.assertEqual(self.graph.allEdges(n), edges)

            self.graph.swapEdgeOrder(n, edges[0], edges[-1])
            edges[0], edges[-1] = edges[-1], edges[0]
            self.assertEqual(list(self.graph.getInOutEdges(n)), edges)
            self.assertEqual(self.graph.allEdges(n), edges)

    def test_edge_extremities_modification(self):

        with self.assertRaises(Exception) as cm:
            self.graph.ends(tlp.edge(NB_EDGES))
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.reverse(tlp.edge(NB_EDGES))
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.setSource(tlp.edge(NB_EDGES), tlp.node())
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.setTarget(tlp.edge(NB_EDGES), tlp.node())
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.setSource(self.graph.getRandomEdge(),
                                 tlp.node(NB_NODES))
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.setTarget(self.graph.getRandomEdge(),
                                 tlp.node(NB_NODES))
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.setEnds(tlp.edge(NB_EDGES), tlp.node(), tlp.node())
            self.assertEquals(
                cm.exception.args[0],
                'Edge with id %s does not belong to graph "%s" (id %s)' %
                (NB_EDGES, self.graph.getName(), self.graph.getId()))

        with self.assertRaises(Exception) as cm:
            self.graph.setEnds(self.graph.getRandomEdge(),
                               tlp.node(NB_NODES), tlp.node(NB_NODES+1))
            self.assertEquals(
                cm.exception.args[0],
                'Node with id %s does not belong to graph "%s" (id %s)' %
                (NB_NODES, self.graph.getName(), self.graph.getId()))

        for e in self.graph.edges():
            ends = self.graph.ends(e)
            self.graph.reverse(e)
            self.assertEqual(self.graph.source(e), ends[1])
            self.assertEqual(self.graph.target(e), ends[0])

            new_src = self.graph.getRandomNode()
            new_tgt = self.graph.getRandomNode()
            self.graph.setSource(e, new_src)
            self.graph.setTarget(e, new_tgt)
            self.assertEqual(self.graph.source(e), new_src)
            self.assertEqual(self.graph.target(e), new_tgt)
            self.assertEqual(self.graph.ends(e), (new_src, new_tgt))

            new_src = self.graph.getRandomNode()
            new_tgt = self.graph.getRandomNode()
            self.graph.setEnds(e, new_src, new_tgt)
            self.assertEqual(self.graph.source(e), new_src)
            self.assertEqual(self.graph.target(e), new_tgt)
            self.assertEqual(self.graph.ends(e), (new_src, new_tgt))



# This file is part of Tulip (http://tulip.labri.fr)
#
# Authors: David Auber and the Tulip development Team
# from LaBRI, University of Bordeaux
#
# Tulip is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Tulip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.

# That plugin imports a graph from a file in the GraphML format
# (see http://graphml.graphdrawing.org/)

from tulip import tlp
import tulipplugins

import xml.sax

# map GraphML node id to Tulip node
idToNode = {}


# XML Sax Content Handler for parsing GraphML files
class TulipGraphMLHandler(xml.sax.ContentHandler):

    def __init__(self, graph):
        xml.sax.ContentHandler.__init__(self)
        self.graph = graph
        self.viewLayout = graph.getLayoutProperty('viewLayout')
        self.viewSize = graph.getSizeProperty('viewSize')
        self.viewColor = graph.getColorProperty('viewColor')
        self.viewLabel = graph.getStringProperty('viewLabel')
        self.attrLabel = ''
        self.attributes = {}
        self.currentNode = None
        self.currentEdge = None
        self.currentAttrId = None
        self.parsingDefault = False

    # try to create a Tulip property based on GraphML attribute type
    def getGraphProperty(self, attrId, attrName, attrType):
        if attrType == 'string':
            return self.graph.getStringProperty(attrName)
        elif attrType == 'boolean':
            return self.graph.getBooleanProperty(attrName)
        elif attrType in ('int', 'long'):
            return self.graph.getIntegerProperty(attrName)
        elif attrType in ('double', 'float'):
            return self.graph.getDoubleProperty(attrName)
        else:
            return None

    def startElement(self, name, attrs):
        # attribute definition
        if name == 'key':
            # only parse standardly defined attributes
            if 'id' in attrs and 'for' in attrs \
                    and attrs['for'] in ('node', 'edge'):
                self.currentAttrId = attrs['id']
                # get attribute name and type
                attrName = attrs['id']
                attrType = 'string'
                if 'attr.name' in attrs:
                    attrName = attrs['attr.name']
                if 'attr.type' in attrs:
                    attrType = attrs['attr.type']
                # get viewLabel corresponding attribute
                if attrName == 'name' and attrType == 'string':
                    self.attrLabel = 'name'
                elif attrName == 'label' and attrType == 'string' \
                        and self.attrLabel != 'name':
                    self.attrLabel = 'label'

                # Create a Tulip property compatible with the attribute type
                self.getGraphProperty(self.currentAttrId, attrName, attrType)
                # save attributes info for later use
                self.attributes[attrs['id']] = {'name': attrName,
                                                'typename': attrType,
                                                'for': attrs['for']}
        # attribute default value
        elif name == 'default':
            self.parsingDefault = True

        # node definition
        elif name == 'node':
            # get node id
            nodeId = attrs['id']
            # create a Tulip node
            n = self.graph.addNode()
            # map id to node
            idToNode[nodeId] = n
            # set the id as the default node label
            self.graph['viewLabel'][n] = nodeId
            # mark current node
            self.currentNode = n

        # edge definition
        elif name == 'edge':
            # get source and target node id
            sourceId = attrs['source']
            targetId = attrs['target']
            # add the edge if the ids are valid and mark it as current
            if sourceId in idToNode and targetId in idToNode:
                self.currentEdge = self.graph.addEdge(idToNode[sourceId],
                                                      idToNode[targetId])

        # attribute definition
        elif name == 'data':
            # if it is a standard one, mark it as current
            if attrs['key'] in self.attributes:
                self.currentAttrId = attrs['key']

    # reset current element parsed
    def endElement(self, name):
        if name == 'node':
            self.currentNode = None
        elif name == 'edge':
            self.currentEdge = None
        elif name == 'data' or name == 'key':
            self.currentAttrId = None
        elif name == 'default':
            self.parsingDefault = False

    # parse attributes values
    def characters(self, content):
        # nothing to parse
        if content == '\n':
            return

        # parsing attribute default value
        if self.currentAttrId:
            # get the name and the typename of the attribute
            attrName = self.attributes[self.currentAttrId]['name']
            attrType = self.attributes[self.currentAttrId]['typename']
            attrFor = self.attributes[self.currentAttrId]['for']

            if self.parsingDefault and attrFor in ('node', 'edge'):
                # try to create a Tulip property compatible with
                # the attribute type
                graphProperty = self.getGraphProperty(self.currentAttrId,
                                                      attrName, attrType)
                if graphProperty:
                    if attrFor == 'node':
                        graphProperty.setAllNodeStringValue(content)
                    else:
                        graphProperty.setAllEdgeStringValue(content)

            # parsing attribute
            else:
                # parse node attribute
                if self.currentNode:
                    # try to parse some standard node visual attributes
                    # (label, layout, size, color)
                    if attrName == self.attrLabel:
                        self.viewLabel[self.currentNode] = content
                    elif attrName in ('x', 'y'):
                        nodeCoord = self.viewLayout[self.currentNode]
                        if attrName == 'x':
                            nodeCoord[0] = float(content)
                        else:
                            nodeCoord[1] = float(content)
                        self.viewLayout[self.currentNode] = nodeCoord
                    elif attrName in ('width', 'height', 'size'):
                        nodeSize = self.viewSize[self.currentNode]
                        size = float(content)
                        if attrName in ('width', 'size'):
                            nodeSize[0] = size
                        if attrName in ('height', 'size'):
                            nodeSize[1] = size
                        self.viewSize[self.currentNode] = nodeSize
                    elif attrName in ('r', 'g', 'b'):
                        nodeColor = self.viewColor[self.currentNode]
                        if attrName == 'r':
                            nodeColor[0] = int(content)
                        elif attrName == 'g':
                            nodeColor[1] = int(content)
                        else:
                            nodeColor[2] = int(content)
                        self.viewColor[self.currentNode] = nodeColor

                    # try to get a Tulip property compatible with
                    # the attribute type
                    graphProperty = self.getGraphProperty(self.currentAttrId,
                                                          attrName, attrType)
                    if graphProperty:
                        # set the property value from its string representation
                        graphProperty.setNodeStringValue(self.currentNode,
                                                         content)

                # same process for parsing edges attributes
                elif self.currentEdge:
                    graphProperty = self.getGraphProperty(self.currentAttrId,
                                                          attrName, attrType)
                    if graphProperty:
                        graphProperty.setEdgeStringValue(self.currentEdge,
                                                         content)

                # graph attribute case
                else:
                    self.graph.setAttribute(attrName, content)


class GraphMLImport(tlp.ImportModule):
    def __init__(self, context):
        tlp.ImportModule.__init__(self, context)
        self.addFileParameter('filename', True, 'The GraphML file to import')

    def fileExtensions(self):
        return ['graphml']

    def importGraph(self):

        # create an XML Sax parser
        parser = xml.sax.make_parser()
        # set our custom content handler
        parser.setContentHandler(TulipGraphMLHandler(self.graph))
        # parse the GraphML file
        parser.parse(open(self.dataSet['filename'], 'rb'))

        return True


pluginDoc = """
<p>Supported extension: graphml</p><p>Imports a graph from a file in the
GraphML format (http://graphml.graphdrawing.org/).
GraphML is a comprehensive and easy-to-use file format for graphs.
It consists of a language core to describe the structural properties
of a graph and a flexible extension mechanism to add application-specific
data.</p>
"""

# The line below does the magic to register the plugin to the plugin database
# and updates the GUI to make it accessible through the menus.
tulipplugins.registerPluginOfGroup(
    'GraphMLImport', 'GraphML', 'Antoine Lambert', '14/05/2017',
    pluginDoc, '1.0', 'File')



# This file is part of Tulip (http://tulip.labri.fr)
#
# Authors: David Auber and the Tulip development Team
# from LaBRI, University of Bordeaux
#
# Tulip is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Tulip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.

from tulip import tlp
import tulipplugins

numpyOk = True

# h3 module requires numpy so we need to check it is available
# in the Python environment
try:
    # https://github.com/buzzfeed/pyh3
    from h3.tree import Tree
except Exception:
    numpyOk = False


class H3Layout(tlp.LayoutAlgorithm):

    def __init__(self, context):
        tlp.LayoutAlgorithm.__init__(self, context)

        self.addFloatParameter(
            'layout scaling',
            'the scale factor to apply to the computed layout', '1000')

    def check(self):

        if not numpyOk:
            return (False,
                    ('Python numpy module is required to execute that '
                     'layout algorithm.\nYou can easily install it with '
                     'the pip tool or your package manager if you are a '
                     'linux user.'))

        if not tlp.ConnectedTest.isConnected(self.graph):
            return (False, 'The graph must be connected')

        return (True, '')

    def run(self):

        # compute a spanning tree of the input graph
        spanningTree = tlp.TreeTest.computeTree(self.graph)

        # get edges to use an input for the H3 layout implementation
        # from buzzfeed and reverse their orientation as it does not
        # use the same direction as in Tulip graphs
        edges = [self.graph.ends(e)[::-1] for e in spanningTree.getEdges()]

        # compute the layout
        tree = Tree(edges)

        # copy result to Tulip layout property
        self.result.setAllEdgeValue([])
        for n in self.graph.getNodes():
            self.result[n] = (tree.nodes[n].coord.x,
                              tree.nodes[n].coord.y,
                              tree.nodes[n].coord.z)

        # apply some scaling factor to the layout to get a correct rendering
        # in Tulip
        scaling = 1000
        if self.dataSet:
            scaling = self.dataSet['layout scaling']

        self.result.scale((scaling, scaling, scaling))

        # cleanup computed spanning tree
        tlp.TreeTest.cleanComputedTree(self.graph, spanningTree)

        return True


pluginDoc = """
Implements the H3 layout technique for drawing large directed graphs as
node-link diagrams in 3D hyperbolic space. That algorithm can lay out much
larger structures than can be handled using traditional techniques
for drawing general graphs because it assumes a hierarchical nature
of the data.
It was first published as:<br/><b> H3: Laying out Large Directed Graphs
in 3D Hyperbolic Space</b>, Tamara Munzner, Proceedings of the 1997 IEEE
Symposium on Information Visualization, Phoenix, AZ, pp 2-10, 1997.<br/>
The implementation in Python (MIT License) has been written by
BuzzFeed engineers (https://github.com/buzzfeed/pyh3).
"""

# The line below does the magic to register the plugin into the plugin database
# and updates the GUI to make it accessible through the menus.
tulipplugins.registerPluginOfGroup(
    'H3Layout', 'H3', 'Antoine Lambert', '30/08/2017',
    pluginDoc, '1.0', 'Hierarchical')



# Make VCProj
#
# October 2012
# Markus Chimani, markus.chimani@uni-jena.de
# Carsten Gutwenger, carsten.gutwenger@cs.tu-dortmund.de
#########################################################

# use Python 3 print function on both Python 2.x and Python 3.x
from __future__ import print_function

import os, sys, fnmatch

# Simple hack for using "configparser" on both Python 2 and 3
if sys.hexversion > 0x03000000:
	import configparser
else:
	import ConfigParser
	configparser = ConfigParser


class stuff:
	def __init__(self, ttag, tpath, tpats):
		self.tag, self.path, self.pats = ttag, tpath, tpats

def bailout(msg):
	print(msg)
	print('Please use the original makeVCProj.config as a template')
	sys.exit()

def loadConfig(sect, key, noError = False ):
	if config.has_option(sect, key):
		v = config.get(sect, key)
		print('   [', sect, ']', key, '=', v)
		return v
	else:
		if noError:
			return None
		else:
			bailout('Option "' + key + '" in section "' + sect + '" is missing')

def checkSolver(name, defSolver, extSolvers):
	if(name == defSolver):
		return True
	for s in extSolvers:
		s = s.strip()
		if(name == s):
			return True
	return False

def getLibs(libsConfig):
	libs = ''
	for str in libsConfig.split(';'):
		str = str.strip();
		if str != '':
			libs += str + ' '
	return libs


#########################################################
# LOAD CONFIGURATION

config = configparser.ConfigParser()

makecfg = 'makeVCProj.config'
for x in sys.argv[1:]:
    if x[:7] == "config=":
        makecfg = x[7:]

print('Loading ' + makecfg + '...')

try:
	config.readfp( open(makecfg))
except IOError:
	bailout(makecfg + ' not found')

if not config.has_section('OGDF'):
	bailout('Section "OGDF" is missing')
if not config.has_section('COIN'):
	bailout('Section "COIN" is missing')

#########################################################
# CONFIGS

# Filenames
filename_vcproj =  loadConfig('OGDF', 'projectFile')
filename_template = loadConfig('OGDF', 'templateFile')
addDefines = ''
addIncludes = ''
addLibs = ''
addLibsWin32 = ''
addLibsWin32 = ''
addLibsX64 = ''
addLibsX64 = ''

# defines for auto generated config_autogen.h header file
config_defines = ''

createDLL = loadConfig('OGDF', 'DLL', True)

memoryManager = loadConfig('OGDF', 'memoryManager', True)
if memoryManager:
	config_defines += '#define ' + memoryManager + '\n'

ogdfOpenMP = 'false'
coinOpenMP = 'false'
ogdfEnableOpenMP = loadConfig('OGDF', 'OpenMP', True)
if ogdfEnableOpenMP and ogdfEnableOpenMP.startswith('t'):
	ogdfOpenMP = 'true'

useCoin = loadConfig('COIN', 'useCoin').startswith('t')
addOsiCpx = False
addOsiGrb = False
addLibsDebugWin32 = ''
addLibsReleaseWin32 = ''
addLibsDebugX64 = ''
addLibsReleaseX64 = ''
if useCoin:
	filename_coin_vcproj =  loadConfig('COIN', 'projectFile')
	filename_coin_template = loadConfig('COIN', 'templateFile')

	coinEnableOpenMP = loadConfig('COIN', 'OpenMP', True)
	if coinEnableOpenMP and coinEnableOpenMP.startswith('t'):
		coinOpenMP = 'true'

	defaultSolver   = loadConfig('COIN', 'defaultSolver')
	externalSolvers = loadConfig('COIN', 'externalSolvers').split(';')
	solverIncludes  = loadConfig('COIN', 'solverIncludes').split(';')

	addOsiCpx = checkSolver('CPX', defaultSolver, externalSolvers)
	addOsiGrb = checkSolver('GRB', defaultSolver, externalSolvers)

	config_defines += '#define USE_COIN\n'
	config_defines += '#define COIN_OSI_' + defaultSolver + '\n'

	for s in externalSolvers:
		s = s.strip();
		if s != '':
			config_defines += '#define OSI_' + s.strip() + '\n'

	for p in solverIncludes:
		p = p.strip()
		if p != '':
			addIncludes += p + ';'

	if createDLL and createDLL.startswith('t'):
		addLibs += 'coin.lib '

		addLibsDebugWin32   += getLibs(loadConfig('COIN', 'solverLibs_win32_debug'))
		addLibsReleaseWin32 += getLibs(loadConfig('COIN', 'solverLibs_win32_release'))
		addLibsDebugX64     += getLibs(loadConfig('COIN', 'solverLibs_x64_debug'))
		addLibsReleaseX64   += getLibs(loadConfig('COIN', 'solverLibs_x64_release'))

linkSectionD32 = ''
linkSectionR32 = ''
linkSectionD64 = ''
linkSectionR64 = ''
if createDLL and createDLL.startswith('t'):
	libraryType = '2'
	config_defines += '#define OGDF_DLL\n'
	addDefines += 'OGDF_INSTALL'
	linkerName = 'VCLinkerTool'
	linkSectionBegin = '				AdditionalDependencies=\"psapi.lib '
	linkSectionEnd = '\"\n\
				AdditionalLibraryDirectories="$(SolutionDir)$(PlatformName)\$(ConfigurationName)"\
				LinkTimeCodeGeneration="0"'

	linkSectionD32 = linkSectionBegin + addLibs + addLibsDebugWin32   + linkSectionEnd
	linkSectionR32 = linkSectionBegin + addLibs + addLibsReleaseWin32 + linkSectionEnd
	linkSectionD64 = linkSectionBegin + addLibs + addLibsDebugX64     + linkSectionEnd
	linkSectionR64 = linkSectionBegin + addLibs + addLibsReleaseX64   + linkSectionEnd

else:
	libraryType = '4'
	linkerName = 'VCLibrarianTool'

print('Generating config_autogen.h ...')
config_autogen = open('include\\ogdf\\internal\\config_autogen.h','w')
config_autogen.write('//\n')
config_autogen.write('// This file has been automatically generated by makeVCProj.py\n')
config_autogen.write('//\n')
config_autogen.write('// Do not change this file manually, instead reconfigure OGDF!\n')
config_autogen.write('\n')
config_autogen.write(config_defines)
config_autogen.close()

addIncludes = addIncludes[:-1]

libraryTypeTag = '<<LIBRARYTYPETAG>>'
defineTag = '<<DEFINETAG>>'
includeTag = '<<INCLUDETAG>>'
linkerNameTag = '<<LINKERNAME>>'
linkTagD32 = '<<LINKTAGD32>>'
linkTagR32 = '<<LINKTAGR32>>'
linkTagD64 = '<<LINKTAGD64>>'
linkTagR64 = '<<LINKTAGR64>>'
openMPTag = '<<OPENMPTAG>>'

# Params are:
# - Tag in template-File
# - Directory to start search & subfilters from
# - File Patterns
cppStuff = stuff( '<<CPPTAG>>', 'src\\ogdf', [ '*.c', '*.cpp' ] )
hStuff = stuff( '<<HTAG>>', 'include\\ogdf', [ '*.h' ] )
hLegacyStuff = stuff( '<<HLEGACYTAG>>', 'include\\ogdf_legacy', [ '*.h' ] )
ogdfStuff = [ cppStuff, hStuff, hLegacyStuff ]

cppCoinStuff = stuff( '<<CPPTAG>>', 'src\\coin', [ '*.c', '*.cpp' ] )
hCoinStuff = stuff( '<<HTAG>>', 'include\\coin', [ '*.h', '*.hpp' ] )
coinStuff = [ cppCoinStuff, hCoinStuff ]


#########################################################
#########################################################
## only code below...

# just the def. nothing happens yet
def Walk( curdir, pats, intro ):
	names = os.listdir( curdir)
	names.sort()

	for name in names:
		# OGDF ignores
		if name.startswith('.') or name.startswith('_') or (name=='legacy' and not includeLegacyCode):
			continue
		# COIN ignores
		if (name == 'OsiCpxSolverInterface.cpp' or name == 'OsiCpxSolverInterface.hpp') and not addOsiCpx:
			continue
		if (name == 'OsiGrbSolverInterface.cpp' or name == 'OsiGrbSolverInterface.hpp') and not addOsiGrb:
			continue

		outpath = curdir + '\\' + name
		fullname = os.path.normpath(outpath)

		if os.path.isdir(fullname) and not os.path.islink(fullname):
			if name != 'abacus' or useCoin == True:
				if Walk( outpath, pats, intro + '<Filter Name="' + name + '">\n'):
					intro = ''
					vcproj.write('</Filter>\n')
		else:
			for pat in pats:
				if fnmatch.fnmatch(name, pat):
					if len(intro)>0:
						vcproj.write(intro)
						intro = ''
					vcproj.write('<File RelativePath="' + outpath + '"> </File>\n')
	return len(intro) == 0

##########################################
## Main below...

print('Generating VCProj...', end='')

includeLegacyCode = 0;
if len(sys.argv)>1 and sys.argv[1]=='legacy':
	includeLegacyCode = 1
	print('(including legacy code)')

vcproj = open(filename_vcproj,'w')
template = open(filename_template)

check = 0
for line in template:
	if check < len(ogdfStuff) and line.find(ogdfStuff[check].tag) > -1:
		if (ogdfStuff[check].tag!='<<HLEGACYTAG>>' or includeLegacyCode):
			Walk(ogdfStuff[check].path, ogdfStuff[check].pats, '')
		check = check + 1
	elif line.find(libraryTypeTag) > -1:
		vcproj.write(line.replace(libraryTypeTag,libraryType,1))
	elif line.find(openMPTag) > -1:
		vcproj.write(line.replace(openMPTag,ogdfOpenMP,1))
	elif line.find(defineTag) > -1:
		vcproj.write(line.replace(defineTag,addDefines,1))
	elif line.find(includeTag) > -1:
		vcproj.write(line.replace(includeTag,addIncludes,1))
	elif line.find(linkerNameTag) > -1:
		vcproj.write(line.replace(linkerNameTag,linkerName,1))
	elif line.find(linkTagD32) > -1:
		vcproj.write(line.replace(linkTagD32,linkSectionD32,1))
	elif line.find(linkTagR32) > -1:
		vcproj.write(line.replace(linkTagR32,linkSectionR32,1))
	elif line.find(linkTagD64) > -1:
		vcproj.write(line.replace(linkTagD64,linkSectionD64,1))
	elif line.find(linkTagR64) > -1:
		vcproj.write(line.replace(linkTagR64,linkSectionR64,1))
	else:
		vcproj.write(line)

template.close()
vcproj.close()

print('done')

if useCoin:
	print ('Generating COIN VCXProj...', end='')

	vcproj = open(filename_coin_vcproj,'w')
	template = open(filename_coin_template)

	check = 0
	for line in template:
		if check < len(coinStuff) and line.find(coinStuff[check].tag) > -1:
			Walk(coinStuff[check].path, coinStuff[check].pats, '')
			check = check + 1
		elif line.find(openMPTag) > -1:
			vcproj.write(line.replace(openMPTag,coinOpenMP,1))
		elif line.find(includeTag) > -1:
			vcproj.write(line.replace(includeTag,addIncludes,1))
		else:
			vcproj.write(line)

	template.close()
	vcproj.close()

	print ('done')



# This file is part of Tulip (http://tulip.labri.fr)
#
# Authors: David Auber and the Tulip development Team
# from LaBRI, University of Bordeaux
#
# Tulip is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3
# of the License, or (at your option) any later version.
#
# Tulip is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.

# Tulip import plugin written in Python for importing the dependencies
# graph of a CMake project
# (target -> target dependencies + target -> external libraries dependencies)

# The plugin use the graphviz output feature of CMake enabling to
# generate dot files representing the different dependencies graphs of
# the project (global + target specific). The global dot file is then
# imported in Tulip through the graphviz import plugin included
# in the framework

from tulip import tlp
import tulipplugins

import tempfile
import subprocess
import os
import shutil

cmakeProjectSourceDirParamName = 'CMake project source dir'
cmakeExecutableParamName = 'CMake executable'
cmakeParametersParamName = 'CMake parameters'

cmakeProjectDotFileName = 'cmake_project'


def executeCommand(command, cwd, stderrfile, pluginProgress):
    # execute the process
    p = subprocess.Popen(command, cwd=cwd, shell=True, stdout=subprocess.PIPE,
                         stderr=open(stderrfile, 'wb'))
    # add some execution feedback trough the plugin progress
    while True:
        line = p.stdout.readline().decode('utf-8')
        line = line.replace('\r\n', '').replace('\n', '')
        if line == '' and p.poll() is not None:
            break
        if pluginProgress:
            pluginProgress.setComment(line)

    # wait for the process to complete
    retval = p.wait()
    return retval


class CMakeDependenciesGraphImport(tlp.ImportModule):
    def __init__(self, context):
        tlp.ImportModule.__init__(self, context)

        self.addDirectoryParameter(
            cmakeProjectSourceDirParamName,
            'The root source directory of the CMake project')

        self.addFileParameter(
            cmakeExecutableParamName, True,
            ('Optional parameter in order to provide the path to the CMake '
             'executable. By default CMake executable path is assumed to be '
             'in your PATH environment variable'), 'cmake')

        self.addStringParameter(
            cmakeParametersParamName,
            ('Optional parameter for providing some parameters to CMake in '
             'order to correctly configure the project'))

    def icon(self):
        return '%s/cmake.png' % os.path.dirname(__file__)

    def importGraph(self):

        # get the CMake project source directory path
        cmakeProjectSourceDir = self.dataSet[cmakeProjectSourceDirParamName]

        # create a temporary directory that we will use as a CMake
        # binary directory
        tmpdir = tempfile.mkdtemp()

        # build the command line to execute cmake and generate the dot files
        command = '%s %s --graphviz=%s %s' % (
            self.dataSet[cmakeExecutableParamName],
            self.dataSet[cmakeParametersParamName],
            cmakeProjectDotFileName,
            cmakeProjectSourceDir
        )

        # create a temporary file in which we will redirect stderr
        stderrfile = tmpdir + '/stderr.txt'

        # execute the CMake project configuration
        retval = executeCommand(command, tmpdir, stderrfile,
                                self.pluginProgress)

        # Special case for MinGW Makefiles as it often fails the first time it
        # is executed due to the presence of sh.exe in the PATH environment
        # variable. It fails the first time CMake is executed but not for the
        # subsequent calls, so relaunch the CMake configuration
        mingw = 'MinGW Makefiles' in self.dataSet[cmakeParametersParamName]
        if retval != 0 and mingw:
            retval = executeCommand(command, tmpdir, stderrfile,
                                    self.pluginProgress)

        # something went wrong
        if retval != 0:
            # read stderr file and transmit error string to the plugin progress
            errors = open(stderrfile, 'rb').read().decode('utf-8')
            if self.pluginProgress:
                self.pluginProgress.setError(errors)
            # remove temporary directory
            shutil.rmtree(tmpdir)
            return False

        # check if CMake has generated the expected dot file
        cmakeProjectDotFilePath = tmpdir + '/' + cmakeProjectDotFileName
        if not os.path.exists(cmakeProjectDotFilePath):
            cmakeProjectDotFilePath = '%s/%s.dot' % (
                tmpdir, cmakeProjectDotFileName)
            if not os.path.exists(cmakeProjectDotFilePath):
                if self.pluginProgress:
                    self.pluginProgress.setError(
                        ('Error : the graphviz file has not been generated '
                         'by CMake'))
                # remove temporary directory
                shutil.rmtree(tmpdir)
                return False

        # import the dot file in Tulip trough the graphviz import plugin
        dotImportParams = tlp.getDefaultPluginParameters('graphviz',
                                                         self.graph)
        dotImportParams['filename'] = cmakeProjectDotFilePath
        tlp.importGraph("graphviz", dotImportParams, self.graph)

        # set some visual attributes to the imported graph
        self.graph['viewSize'].setAllNodeValue((1, 1, 1))
        self.graph['viewShape'].setAllNodeValue(tlp.NodeShape.Circle)

        # apply a force directed algorithm to draw the graph
        self.graph.applyLayoutAlgorithm(
            'Fast Multipole Multilevel Embedder (OGDF)')

        # remove temporary directory
        shutil.rmtree(tmpdir)

        # done
        return True


# The line below does the magic to register the plugin to the plugin database
# and updates the GUI to make it accessible through the menus.
tulipplugins.registerPluginOfGroup(
    'CMakeDependenciesGraphImport', 'CMake dependencies graph',
    'Antoine Lambert', '24/11/2015',
    "Import the targets dependencies graph of a CMake project",
    '1.0', 'Software')
